from retro.graph import Vertex, Edge, Graph
from retro.errors import TerminalTooSmall

identity = lambda x: x

def get_agent_character(agent, position):
    """Returns the character to display for an agent at the given board position.
    For agents with a size and a multi-row character grid, returns the
    appropriate character for that specific cell.
    """
    char = agent.character
    if isinstance(char, str) or not hasattr(agent, 'size'):
        return char
    ax, ay = agent.position
    x, y = position
    dx, dy = x - ax, y - ay
    rows = list(char)
    if dy >= len(rows):
        return ' '
    row = rows[dy]
    if dx >= len(row):
        return ' '
    return row[dx]

def vector_add(vec0, vec1):
    "Adds two vectors."
    x0, y0 = vec0
    x1, y1 = vec1
    return (x0 + x1, y0 + y1)

class View:
    BORDER_X = 2
    BORDER_Y = 3
    STATE_HEIGHT = 5
    DEBUG_WIDTH = 60

    def __init__(self, terminal, color='white_on_black'):
        self.terminal = terminal
        self.terminal_size = (self.terminal.width, self.terminal.height)
        self.color = color
        self.initial_render = True

    def render(self, game):
        if self.initial_render or self.terminal_size_changed():
            self.terminal_size = (self.terminal.width, self.terminal.height)
            self.render_layout(game)
        if self.initial_render or game.state.changed:
            self.render_state(game)
        if game.debug:
            self.render_debug_log(game)
        prior_board_view = self.get_board_view(
            game.prior_agent_positions,
            game.prior_view_position,
            game.view_size
        )
        board_view = self.get_board_view(
            game.agent_positions,
            game.view_position,
            game.view_size
        )
        diff = self.get_board_view_diff(prior_board_view, board_view)
        self.render_board_view(diff, game)
        self.initial_render = False

    def get_board_view(self, agent_positions, view_position, view_size):
        """Returns a dict of view-coordinate position -> colored_character,
        containing only the agents visible within the viewport.
        """
        vox, voy = view_position
        vx, vy = view_size
        board_view = {}
        for position, agents in agent_positions.items():
            x, y = position
            if vox <= x < vox + vx and voy <= y < voy + vy:
                top = max(agents, key=lambda a: getattr(a, 'z', 0) or 0)
                color = self.get_color(top.color) if hasattr(top, 'color') else identity
                board_view[(x - vox, y - voy)] = color(get_agent_character(top, (x, y)))
        return board_view

    def get_board_view_diff(self, board_view_0, board_view_1):
        """Returns a dict of position -> colored_character needed to transform
        board_view_0 to board_view_1.
        """
        diff = {}
        positions = set(board_view_0.keys()).union(board_view_1.keys())
        for p in positions:
            if p not in board_view_0:
                diff[p] = board_view_1[p]
            elif p not in board_view_1:
                diff[p] = self.get_color(self.color)(' ')
            elif board_view_0[p] != board_view_1[p]:
                diff[p] = board_view_1[p]
        return diff

    def render_board_view(self, board_view, game):
        origin = self.get_view_origin_coords(game)
        for position, colored_character in board_view.items():
            x, y = vector_add(origin, position)
            print(self.terminal.move_xy(x, y) + colored_character)

    def render_layout(self, game):
        bw, bh = game.view_size
        self.check_terminal_size(game)
        self.clear_screen()
        layout_graph = self.get_layout_graph(game)
        layout_graph.render(self.terminal)

    def clear_screen(self):
        print(self.terminal.home + self.get_color(self.color) + self.terminal.clear)

    def get_color(self, color_string):
        if not hasattr(self.terminal, color_string):
            msg = (
                f"{color_string} is not a supported color."
                "See https://blessed.readthedocs.io/en/latest/colors.html"
            )
            raise ValueError(msg)
        return getattr(self.terminal, color_string)

    def render_state(self, game):
        vw, vh = game.view_size
        ox, oy = self.get_state_origin_coords(game)
        color = self.get_color(self.color)
        for i, key in enumerate(sorted(game.state.keys())):
            msg = f"{key}: {game.state[key]}"[:vw]
            print(self.terminal.move_xy(ox, oy + i) + color(msg))

    def render_debug_log(self, game):
        vw, vh = game.view_size
        debug_height = vh + self.STATE_HEIGHT
        ox, oy = self.get_debug_origin_coords(game)
        color = self.get_color(self.color)
        for i, (turn_number, message) in enumerate(game.log_messages[-debug_height:]):
            msg = f"{turn_number}. {message}"[:self.DEBUG_WIDTH - 1].ljust(self.DEBUG_WIDTH - 1)
            print(self.terminal.move_xy(ox, oy + i) + color(msg))

    def get_layout_graph(self, game):
        vw, vh = game.view_size
        sh = self.STATE_HEIGHT
        ox, oy = self.get_view_origin_coords(game)

        vertices = [
            Vertex(ox - 1, oy - 1), 
            Vertex(ox + vw, oy - 1),
            Vertex(ox + vw, oy + vh),
            Vertex(ox + vw, oy + vh + sh),
            Vertex(ox - 1, oy + vh + sh),
            Vertex(ox - 1, oy + vh)
        ]
        edges = [
            Edge(vertices[0], vertices[1]),
            Edge(vertices[1], vertices[2]),
            Edge(vertices[2], vertices[3]),
            Edge(vertices[3], vertices[4]),
            Edge(vertices[4], vertices[5]),
            Edge(vertices[5], vertices[0]),
            Edge(vertices[5], vertices[2]),
        ]
        graph = Graph(vertices, edges)
        if game.debug:
            dw = self.DEBUG_WIDTH
            graph.vertices.append(Vertex(ox + vw + dw, oy - 1))
            graph.vertices.append(Vertex(ox + vw + dw, oy + vh + sh))
            graph.edges.append(Edge(graph.vertices[1], graph.vertices[6]))
            graph.edges.append(Edge(graph.vertices[6], graph.vertices[7]))
            graph.edges.append(Edge(graph.vertices[3], graph.vertices[7]))
        return graph

    def terminal_size_changed(self):
        return self.terminal_size != (self.terminal.width, self.terminal.height)

    def check_terminal_size(self, game):
        vw, vh = game.view_size
        width_needed = vw + self.BORDER_X
        height_needed = vh + self.BORDER_Y + self.STATE_HEIGHT
        if self.terminal.width < width_needed:
            raise TerminalTooSmall(width=self.terminal.width, width_needed=width_needed)
        elif self.terminal.height < height_needed:
            raise TerminalTooSmall(height=self.terminal.height, height_needed=height_needed)

    def view_origin(self, game):
        x, y = self.get_view_origin_coords(game)
        return self.terminal.move_xy(x, y)

    def get_view_origin_coords(self, game):
        vw, vh = game.view_size
        margin_top = (self.terminal.height - vh - self.BORDER_Y) // 2
        if game.debug:
            margin_left = (self.terminal.width - vw - self.DEBUG_WIDTH - self.BORDER_X) // 2
        else:
            margin_left = (self.terminal.width - vw - self.BORDER_X) // 2
        return margin_left, margin_top

    def get_state_origin_coords(self, game):
        vw, vh = game.view_size
        ox, oy = self.get_view_origin_coords(game)
        return ox, oy + vh + 1

    def get_debug_origin_coords(self, game):
        vw, vh = game.view_size
        ox, oy = self.get_view_origin_coords(game)
        return ox + vw + 1, oy





class Agent:
    """Represents a character in the game. To create an Agent, define a new
    class with some of the attributes and methods below. You may change any of
    the Agent's attributes at any time, and the result will immediately be 
    visible in the game.

    After you create your Agents, add them to the ``Game``, either when it is created
    or using ``Game.add_agent`` later on. Then the Game will take care of calling 
    the Agent's methods at the appropriate times. 

    Attributes:
        position: (Required) The agent's ``(int, int)`` position. The agent
            will only be displayed is this position is on the board.
        character: (Required unless display is ``False``.) A one-character string 
            which will be displayed at the Agent's position on the game board. 
        name: (Optional) If an agent has a name, it must be unique within the game. 
            Agent names can be used to look up agents with 
            :py:meth:`retro.game.Game.get_agent_by_name`. 
        color (str): (Optional) The agent's color. 
            `Available colors <https://blessed.readthedocs.io/en/latest/colors.html>`_.
        display: (Optional) When ``False``, the Agent will not be displayed on the 
            board. This is useful when you want to create an agent which will be displayed
            later, or when you want to create an agent which acts on the Game indirectly, 
            for example by spawning other Agents. Defaults to True.
        z: (Optional) When multiple Agents have the same position on the board, the
            Agent with the highest ``z`` value will be displayed.
            The Game is played on a two-dimensional (x, y) board, but you can think of
            ``z`` as a third "up" dimension. Defaults to 0.
        size: (Optional) A ``(width, height)`` tuple declaring how many cells this
            agent occupies on the board. When set, ``character`` may be a list of
            strings, one per row. Defaults to ``(1, 1)``.
    """
    character = "*"
    position = (0, 0)
    name = "agent"
    color = "white_on_black"
    display = True
    z = 0

    def play_turn(self, game):
        """If an Agent has this method, it will be called once 
        each turn. 

        Arguments: 
            game (Game): The game which is currently being played will be 
                passed to the Agent, in case it needs to check anything about
                the game or make any changes. 
        """
        pass

    def handle_keystroke(self, keystroke, game):
        """If an Agent has a this method, it will be called every
        time a key is pressed in the game.

        Arguments: 
            keystroke (blessed.keyboard.Keystroke): The key which was pressed. You can 
                compare a Keystroke with a string (e.g. ``if keystroke == 'q'``) to check 
                whether it is a regular letter, number, or symbol on the keyboard. You can 
                check special keys using the keystroke's name 
                (e.g. ``if keystroke.name == "KEY_RIGHT"``). Run your game in debug mode to 
                see the names of keystrokes. 
            game (Game): The game which is currently being played will be 
                passed to the Agent, in case it needs to check anything about
                the game or make any changes. 
            
        """
        pass

class ArrowKeyAgent:
    """A simple agent which can be moved around with the arrow keys.
    """
    name = "ArrowKeyAgent"
    character = "*"
    position = (0,0)
    display = True
    z = 0

    def play_turn(self, game):
        pass

    def handle_keystroke(self, keystroke, game):
        """Moves the agent's position if the keystroke is one of the arrow keys.
        One by one, checks the keystroke's name against each arrow key. 
        Then uses :py:meth:`try_to_move` to check whether the move is on the 
        game's board before moving.
        """
        x, y = self.position
        if keystroke.name == "KEY_RIGHT":
            self.try_to_move((x + 1, y), game)
        elif keystroke.name == "KEY_UP":
            self.try_to_move((x, y - 1), game)
        elif keystroke.name == "KEY_LEFT":
            self.try_to_move((x - 1, y), game)
        elif keystroke.name == "KEY_DOWN":
            self.try_to_move((x, y + 1), game)

    def try_to_move(self, position, game):
        """Moves to the position if it is on the game board. 
        """
        if game.on_board(position):
            self.position = position
            game.log(f"Position: {self.position}")

class CenterViewAgent:
    """An agent which centers the game view on the agent.
    If margin is an integer, then the view will only re-center
    when the agent is within ``margin`` spaces of the edge of the view.
    If margin is None (default), the view re-centers every turn.

    Attributes:
        margin (int or None): Distance from the viewport edge that triggers
            re-centering. Defaults to None (always re-center).
    """
    margin = None

    def play_turn(self, game):
        super().play_turn(game)
        if self.needs_to_center_view(game):
            self.center_view(game)

    def center_view(self, game):
        """Centers the game view on this agent.

        Arguments:
            game (Game): The current game.
        """
        vw, vh = game.view_size
        bw, bh = game.board_size
        x, y = self.position
        new_x = max(0, min(x - vw // 2, bw - vw))
        new_y = max(0, min(y - vh // 2, bh - vh))
        game.view_position = (new_x, new_y)

    def needs_to_center_view(self, game):
        """Returns True if the view should be re-centered this turn.

        Arguments:
            game (Game): The current game.
        """
        if self.margin is None:
            return True
        return self.distance_to_edge_of_view(game) <= self.margin

    def distance_to_edge_of_view(self, game):
        """Returns the shortest distance from this agent to any edge of the
        current view.

        Arguments:
            game (Game): The current game.
        """
        x, y = self.position
        vox, voy = game.view_position
        vw, vh = game.view_size
        return min(x - vox, vox + vw - 1 - x, y - voy, voy + vh - 1 - y)


class Tombstone:
    """A placeholder for a missing agent.
    """
    def __init__(self, position, color):
        self.position = position
        self.color = color

    character = ' '



from retro.examples.beast.helpers import add, distance, get_occupant
from random import random, choice

class Beast:
    """Represents a beast, coming after the player.
    """
    character = "H"
    color = "red"
    probability_of_moving = 0.03
    probability_of_random_move = 0.2
    deadly = True

    def __init__(self, position):
        self.position = position

    def handle_push(self, vector, game):
        future_position = add(self.position, vector)
        on_board = game.on_board(future_position)
        obstacle = get_occupant(game, future_position)
        if obstacle or not on_board:
            self.die(game)
            return True
        else:
            return False

    def play_turn(self, game):
        if self.should_move():
            possible_moves = []
            for position in self.get_adjacent_positions():
                if game.is_empty(position) and game.on_board(position):
                    possible_moves.append(position)
            if possible_moves:
                if self.should_move_randomly():
                    self.position = choice(possible_moves)
                else:
                    self.position = self.choose_best_move(possible_moves, game)
            player = game.get_agent_by_name("player")
            if player.position == self.position:
                player.die()

    def get_adjacent_positions(self):
        """Returns a list of all adjacent positions, including diagonals
        """
        positions = []
        for i in [-1, 0, 1]:
            for j in [-1, 0, 1]:
                if i or j:
                    positions.append(add(self.position, (i, j)))
        return positions

    def should_move(self):
        return random() < self.probability_of_moving

    def should_move_randomly(self):
        return random() < self.probability_of_random_move

    def choose_best_move(self, possible_moves, game):
        player = game.get_agent_by_name("player")
        move_distances = [[distance(player.position, move), move] for move in possible_moves]
        shortest_distance, best_move = sorted(move_distances)[0]
        return best_move

    def die(self, game):
        game.remove_agent(self)
        game.num_beasts -= 1
        if game.num_beasts == 0:
            game.state["message"] = "You win!"
            game.end()




def validate_agent(agent):
    if hasattr(agent, "name"):
        validate_agent_name(agent.name)
    name = getattr(agent, "name", agent.__class__.__name__)
    if getattr(agent, 'display', True):
        validate_position(agent.position)
        if not hasattr(agent, "character"):
            raise ValueError(f"Agent {name} must have a character")
        if not isinstance(getattr(agent, 'z', 0), int):
            raise ValueError(f"Agent {name} has invalid z value {agent.z}. z-values must be ints")
        validate_agent_size(agent)
    return agent

def validate_agent_size(agent):
    from retro.errors import AgentCharacterTooLarge
    size = getattr(agent, 'size', None)
    if size is None:
        return
    name = getattr(agent, "name", agent.__class__.__name__)
    if (not isinstance(size, tuple) or len(size) != 2
            or not isinstance(size[0], int) or not isinstance(size[1], int)
            or size[0] < 1 or size[1] < 1):
        raise ValueError(
            f"Agent {name} has invalid size {size}. "
            f"size must be a tuple of two positive integers, e.g. (3, 2)."
        )
    char = agent.character
    if not isinstance(char, str):
        rows = list(char)
        char_height = len(rows)
        char_width = max((len(row) for row in rows), default=0)
        w, h = size
        if char_height > h or char_width > w:
            raise AgentCharacterTooLarge(agent, size, (char_width, char_height))

def validate_state(state):
    if not isinstance(state, dict):
        raise TypeError(f"State is {type(state)}, but must be a dict.")
    for key, value in state.items():
        if is_mutable(value):
            raise ValueError(f"State must be immutable, but state[{key}] is {value}")
    return state

def validate_agent_name(name):
    if not isinstance(name, str):
        raise TypeError(f"Agent names must be strings")
    return name

def validate_position(position):
    if not isinstance(position, tuple):
        raise TypeError(f"Position is {type(position)}, but must be a tuple.")
    if not len(position) == 2:
        raise ValueError(f"Position is {position}. Must be a tuple of two integers.")
    if not isinstance(position[0], int) and isinstance(position[1], int):
        raise TypeError(f"Position is {position}. Must be a tuple of two integers.")
    return position
    
def is_mutable(obj):
    if isinstance(obj, (int, float, bool, str, None)):
        return False
    elif isinstance(obj, tuple):
        return all(is_mutable(element) for element in obj)
    else:
        return True








from retro.examples.beast.helpers import add, get_occupant

class Block:
    """Represents a static block. It just sits there.
    """
    character = "█"
    color = "green4"
    deadly = False

    def __init__(self, position):
        self.position = position

    def handle_push(self, vector, game):
        """Responds to a push in the direction of vector. 
        Returns True when the push succeeds in creating
        empty space.
        """
        future_position = add(self.position, vector)
        on_board = game.on_board(future_position)
        obstacle = get_occupant(game, future_position)
        if obstacle:
            success = obstacle.handle_push(vector, game)
        else:
            success = on_board
        if success:
            self.position = future_position
        return success



from random import randint
from retro.game import Game

class Apple:
    """An agent representing the Apple. 
    Note how Apple doesn't have ``play_turn`` or 
    ``handle_keystroke`` methods: the Apple doesn't need to do 
    anything in this game. It just sits there waiting to get 
    eaten.

    Attributes:
        name: "Apple"
        character: '@'
        color: "red_on_black" (`Here's documentation on how colors 
            work <https://blessed.readthedocs.io/en/latest/colors.html>`_
        position: (0, 0). The Apple will choose a random position 
            as soon as the game starts, but it needs an initial 
            position to be assigned.

    """
    name = "Apple"
    character = '@'
    color = "red_on_black"
    position = (0, 0)

    def relocate(self, game):
        """Sets position to a random empty position. This method is 
        called whenever the snake's head touches the apple. 

        Arguments: 
            game (Game): The current game.
        """
        self.position = self.random_empty_position(game)

    def random_empty_position(self, game):
        """Returns a randomly-selected empty position. Uses a very 
        simple algorithm: Get the game's board size, choose a 
        random x-value between 0 and the board width, and choose 
        a random y-value between 0 and the board height. Now use 
        the game to check whether any Agents are occupying this 
        position. If so, keep randomly choosing a new position 
        until the position is empty.
        """
        bw, bh = game.board_size
        occupied_positions = game.get_agents_by_position()
        while True:
            position = (randint(0, bw-1), randint(0, bh-1))
            if position not in occupied_positions:
                return position

class SnakeHead:
    """An Agent representing the snake's head. When the game starts, you control
    the snake head using the arrow keys. The SnakeHead always has a direction, and 
    will keep moving in that direction every turn. When you press an arrow key, 
    you change the SnakeHead's direction. 

    Attributes:
        name: "Snake head"
        position: (0,0) 
        character: ``'v'`` Depending on the snake head's direction, its character
            changes to ``'<'``, ``'^'``, ``'>'``, or ``'v'``.
        next_segment: Initially ``None``, this is a reference to a SnakeBodySegment.
        growing: When set to True, the snake will grow a new segment on its next move.
    """
    RIGHT = (1, 0)
    UP = (0, -1)
    LEFT = (-1, 0)
    DOWN = (0, 1)
    name = "Snake head"
    position = (0, 0)
    direction = DOWN
    character = 'v'
    next_segment = None
    growing = False

    def play_turn(self, game):
        """On each turn, the snake head uses its position and direction to figure out
        its next position. If the snake head is able to move there (it's on the board and 
        not occuppied by part of the snake's body), it moves. 

        Then, if the snake head is on the Apple, the Apple moves to a new random position
        and ``growing`` is set to True.

        Now we need to deal with two situations. First, if ``next_segment`` is not None, there is 
        a SnakeBodySegment attached to the head. We need the body to follow the head, 
        so we call ``self.next_segment.move``, passing the head's old position 
        (this will be the body's new position), a reference to the game, and a value for 
        ``growing``. If the snake needs to grow, we need to pass this information along
        the body until it reaches the tail--this is where the next segment will be attached. 

        If there is no ``next_segment`` but ``self.growing`` is True, it's time to add 
        a body! We set ``self.next_segment`` to a  new SnakeBodySegment, set its 
        position to the head's old position, and add it to the game. We also add 1 to the 
        game's score.
        """
        x, y = self.position
        dx, dy = self.direction
        if self.can_move((x+dx, y+dy), game):
            self.position = (x+dx, y+dy)
            if self.is_on_apple(self.position, game):
                apple = game.get_agent_by_name("Apple")
                apple.relocate(game)
                self.growing = True
            if self.next_segment:
                self.next_segment.move((x, y), game, growing=self.growing)
            elif self.growing:
                self.next_segment = SnakeBodySegment(1, (x, y))
                game.add_agent(self.next_segment)
                game.state['score'] += 1
            self.growing = False

    def handle_keystroke(self, keystroke, game):
        """Checks whether one of the arrow keys has been pressed. 
        If so, sets the SnakeHead's direction and character.
        """
        x, y = self.position
        if keystroke.name == "KEY_RIGHT":
            self.direction = self.RIGHT
            self.character = '>'
        elif keystroke.name == "KEY_UP":
            self.direction = self.UP
            self.character = '^'
        elif keystroke.name == "KEY_LEFT":
            self.direction = self.LEFT
            self.character = '<'
        elif keystroke.name == "KEY_DOWN":
            self.direction = self.DOWN
            self.character = 'v'

    def can_move(self, position, game):
        on_board = game.on_board(position)
        empty = game.is_empty(position)
        on_apple = self.is_on_apple(position, game)
        return on_board and (empty or on_apple)

    def is_on_apple(self, position, game):
        apple = game.get_agent_by_name("Apple")
        return apple.position == position

class SnakeBodySegment:
    """Finally, we need an Agent for the snake's body segments. 
    SnakeBodySegment doesn't have ``play_turn`` or ``handle_keystroke`` methods because
    it never does anything on its own. It only moves when the SnakeHead, or the previous
    segment, tells it to move. 

    Arguments: 
        segment_id (int): Keeps track of how far back this segment is from the head.
            This is used to give the segment a unique name, and also to keep track 
            of how many points the player earns for eating the next apple. 
        position (int, int): The initial position.

    Attributes:
        character: '*'
        next_segment: Initially ``None``, this is a reference to a SnakeBodySegment
            when this segment is not the last one in the snake's body.
            
    """
    character = '*'
    next_segment = None

    def __init__(self, segment_id, position):
        self.segment_id = segment_id
        self.name = f"Snake body segment {segment_id}"
        self.position = position

    def move(self, new_position, game, growing=False):
        """When SnakeHead moves, it sets off a chain reaction, moving all its 
        body segments. Whenever the head or a body segment has another segment 
        (``next_segment``), it calls that segment's ``move`` method.

        This method updates the SnakeBodySegment's position. Then, if 
        ``self.next_segment`` is not None, calls that segment's ``move`` method. 
        If there is no next segment and ``growing`` is True, then we set
        ``self.next_segment`` to a new SnakeBodySegment in this segment's old
        position, and update the game's score.

        Arguments: 
            new_position (int, int): The new position.
            game (Game): A reference to the current game.
            growing (bool): (Default False) When True, the snake needs to 
                add a new segment. 
        """
        old_position = self.position
        self.position = new_position
        if self.next_segment:
            self.next_segment.move(old_position, game, growing=growing)
        elif growing:
            self.next_segment = SnakeBodySegment(self.segment_id + 1, old_position)
            game.add_agent(self.next_segment)
            game.state['score'] += self.segment_id + 1

if __name__ == '__main__':
    head = SnakeHead()
    apple = Apple()
    game = Game([head, apple], {'score': 0}, board_size=(32, 16), framerate=12)
    apple.relocate(game)
    game.play()




class ChangeDict(dict):
    """A subclass of dict which keeps track of whether it has been changed.
    """
    def __init__(self, *args, **kwargs):
        self.changed = False
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if not (self.get(key) == value):
            self.changed = True
        super().__setitem__(key, value)

        



class Token:
    """A token represents a piece on the board in the virus game. 
    Tokens never move. A token's character and color are determined
    by which player it belongs to and whether it is alive or dead. 
    """
    def __init__(self, position, first_player=True, alive=True):
        self.position = position
        self.first_player = first_player
        self.alive = alive
        self.set_appearance()

    def set_appearance(self):
        self.set_color()
        self.set_character()

    def set_color(self):
        if self.first_player: 
            self.color = "red"
        else:
            self.color = "blue"

    def set_character(self):
        if self.alive: 
            self.character = "O"
        else:
            self.character = "*"



from random import shuffle
from retro.examples.beast.agents.player import Player
from retro.examples.beast.agents.beast import Beast
from retro.examples.beast.agents.block import Block

class Board:
    """The Board creates the agents needed at the beginning of the game, 
    and assigns them all their positions. Board itself is not an agent, 
    and has no role once the game starts.
    """
    def __init__(self, width, height, block_density=0.3, num_beasts=10):
        self.width = width
        self.height = height
        self.block_density = block_density
        self.num_blocks = round(width * height * block_density)
        self.num_empty_spaces = width * height- self.num_blocks
        self.num_beasts = num_beasts
        self.validate()

    def validate(self):
        """Checks that the inputs are valid.
        """
        if self.block_density < 0 or self.block_density > 1:
            raise ValueError("block density must be between 0 and 1.")
        if self.num_empty_spaces < self.num_beasts + 1:
            raise ValueError("Not enough space on the board.")

    def get_agents(self):
        """Returns a list of agents, all initialized in their starting positions.
        """
        positions = self.get_all_positions()
        shuffle(positions)

        player_position = positions[0]
        beast_positions = positions[1:self.num_beasts + 1]
        block_positions = positions[-self.num_blocks:]

        player = [Player(player_position)]
        beasts = [Beast(pos) for pos in beast_positions]
        blocks = [Block(pos) for pos in block_positions]
        return player + beasts + blocks

    def get_all_positions(self):
        """Returns a list of all positions which are not on the edge of the board
        """
        positions = []
        for i in range(self.width):
            for j in range(self.height):
                positions.append((i, j))
        return positions



from retro.agent import ArrowKeyAgent, CenterViewAgent

class Player(ArrowKeyAgent, CenterViewAgent):
    character = "*"
    color = "red_on_black"

    def __init__(self, position):
        self.position = position



from retro.game import Game
from retro.examples.beast.board import Board

WIDTH = 40
HEIGHT = 20
NUM_BEASTS = 10

board = Board(WIDTH, HEIGHT, num_beasts=NUM_BEASTS)
state = {}
game = Game(board.get_agents(), state, board_size=(WIDTH, HEIGHT))
game.num_beasts = NUM_BEASTS
game.play()



from random import randint
from retro.game import Game

HEIGHT = 25
WIDTH = 25

class Spaceship:
    """A player-controlled agent which moves left and right, dodging asteroids.
    Spaceship is a pretty simple class. The ship's character is ``^``, and 
    its position starts at the bottom center of the screen. 
    """
    name = "ship"
    character = '^'
    position = (WIDTH // 2, HEIGHT - 1)
    color = "black_on_skyblue1"

    def handle_keystroke(self, keystroke, game):
        """When the 
        left or arrow key is pressed, it moves left or right. If the ship's 
        new position is empty, it moves to that position. If the new position 
        is occupied (by an asteroid!) the game ends. 
        """
        x, y = self.position
        if keystroke.name in ("KEY_LEFT", "KEY_RIGHT"):
            if keystroke.name == "KEY_LEFT": 
                new_position = (x - 1, y)
            else: 
                new_position = (x + 1, y)
            if game.on_board(new_position):
                if game.is_empty(new_position):
                    self.position = new_position
                else:
                    self.explode()
                    game.end()

    def explode(self):
        """Sets the ship's character to ``*`` and its color to red.
        """
        self.color = "crimson_on_skyblue1"
        self.character = '*'

class Asteroid:
    """When Asteroids are spawned, they fall down the screen until they
    reach the bottom row and are removed. 
    An Asteroid's position is set when it is created.
    Whenever an asteroid moves, it 
    checks whether it has it the ship. 
    """
    character = 'O'
    color = "deepskyblue1_on_skyblue1"

    def __init__(self, position):
        self.position = position

    def play_turn(self, game):
        """Nothing happens unless
        ``game.turn_number`` is divisible by 2. The result is that asteroids 
        only move on even-numbered turns. If the asteroid is at the bottom of 
        the screen, it has run its course and should be removed from the game. 
        Otherwise, the asteroid's new position is one space down from its old 
        position. If the asteroid's new position is the same as the ship's 
        position, the game ends. 
        """
        if game.turn_number % 2 == 0: 
            self.set_color()
            x, y = self.position
            if y == HEIGHT - 1: 
                game.remove_agent(self)
            else:
                ship = game.get_agent_by_name('ship')
                new_position = (x, y + 1)
                if new_position == ship.position:
                    ship.explode()
                    game.end()
                else:
                    self.position = new_position

    def set_color(self):
        """To add to the game's drama, asteroids gradually become visible as they
        fall down the screen. This method calculates the ratio of the asteroid's 
        position compared to the screen height--0 is the top of the screen and 1 is
        the bottom ot the screen. Then sets the asteroid's color depending on the
        ratio. (`Available colors <https://blessed.readthedocs.io/en/latest/colors.html>`_)
        """
        x, y = self.position
        ratio = y / HEIGHT
        if ratio < 0.2: 
            self.color = "deepskyblue1_on_skyblue1"
        elif ratio < 0.4: 
            self.color = "deepskyblue2_on_skyblue1"
        elif ratio < 0.6:
            self.color = "deepskyblue3_on_skyblue1"
        else:
            self.color = "deepskyblue4_on_skyblue1"

class AsteroidSpawner:
    """An agent which is not displayed on the board, but which constantly spawns
    asteroids.
    """
    display = False

    def play_turn(self, game):
        """Adds 1 to the game score and then uses 
        :py:meth:`~retro.examples.nav.should_spawn_asteroid` to decide whether to 
        spawn an asteroid. When :py:meth:`~retro.examples.nav.should_spawn_asteroid` 
        comes back ``True``, creates a new instance of 
        :py:class:`~retro.examples.nav.Asteroid` at a random position along the 
        top of the screen and adds the asteroid to the game. 
        """
        game.state['score'] += 1
        if self.should_spawn_asteroid(game.turn_number):
            asteroid = Asteroid((randint(0, WIDTH - 1), 0))
            game.add_agent(asteroid)

    def should_spawn_asteroid(self, turn_number):
        """Decides whether to spawn an asteroid.
        Uses a simple but effective algorithm to make the game get
        progressively more difficult: choose a random number and return 
        ``True`` if the number is less than the current turn number. At 
        the beginning of the game, few asteroids will be spawned. As the 
        turn number climbs toward 1000, asteroids are spawned almost 
        every turn. 

        Arguments:
            turn_number (int): The current turn in the game.
        """
        return randint(0, 1000) < turn_number

if __name__ == '__main__':
    ship = Spaceship()
    spawner = AsteroidSpawner()
    game = Game(
        [ship, spawner],
        {"score": 0},
        board_size=(WIDTH, HEIGHT),
        color="deepskyblue4_on_skyblue1",
    )
    game.play()




from retro.examples.beast.helpers import add, get_occupant

direction_vectors = {
    "KEY_RIGHT": (1, 0),
    "KEY_UP": (0, -1),
    "KEY_LEFT": (-1, 0),
    "KEY_DOWN": (0, 1),
}

class Player:
    character = "*"
    color = "white"
    name = "player"

    def __init__(self, position):
        self.position = position

    def handle_keystroke(self, keystroke, game):
        if keystroke.name in direction_vectors:
            vector = direction_vectors[keystroke.name]
            self.try_to_move(vector, game)

    def try_to_move(self, vector, game):
        """Tries to move the player in the direction of vector.
        If the space is empty and it's on the board, then the move succeeds. 
        If the space is occupied, then if the occupant can be pushed, it gets
        pushed and the move succeeds. Otherwise, the move fails. 
        """
        future_position = add(self.position, vector)
        on_board = game.on_board(future_position)
        obstacle = get_occupant(game, future_position)
        if obstacle:
            if obstacle.deadly:
                self.die(game)
            elif obstacle.handle_push(vector, game):
                self.position = future_position
        elif on_board:
            self.position = future_position

    def die(self, game):
        self.color = "black_on_red"
        game.state["message"] = "The beasties win!"
        game.end()




from retro.game import Game
from retro.examples.virus.token import Token

size = 12
initial_tokens = [
    Token((0, 0), first_player=True),
    Token((size-1, size-1), first_player=False),
]
game = Game(initial_tokens, {}, (size, size))
game.play()



from retro.errors import GraphError

class Graph:
    def __init__(self, vertices=None, edges=None):
        self.vertices = vertices or []
        self.edges = edges or []

    def __str__(self):
        return '\n'.join(str(e) for e in self.edges)

    def get_or_create_vertex(self, x, y):
        for v in self.vertices:
            if x == v.x and y == v.y:
                return v
        for e in self.edges:
            if e.crosses(x, y):
                return self.split_edge(e, x, y)
        new_vertex = Vertex(x, y)
        self.vertices.append(new_vertex)
        return new_vertex

    def get_or_create_edge(self, x0, y0, x1, y1):
        v0 = self.get_or_create_vertex(x0, y0)
        v1 = self.get_or_create_vertex(x1, y1)
        new_edge = Edge(v0, v1)
        for e in self.edges:
            if e == new_edge:
                new_edge.remove()
                return e
        self.edges.append(new_edge)
        return new_edge

    def split_edge(self, edge, x, y):
        """
        Splits an edge by inserting a new vertex along the edge. 
        """
        if not edge.crosses(x, y):
            raise GraphError(f"Can't split edge {edge} at ({x}, {y})")
        self.remove_edge(edge)
        v = Vertex(x, y)
        self.vertices.append(v)
        self.edges.append(Edge(edge.begin, v))
        self.edges.append(Edge(v, edge.end))
        return v

    def remove_edge(self, edge):
        if edge not in self.edges:
            raise GraphError(f"Edge {edge} is not in the graph")
        self.edges.remove(edge)
        edge.begin.edges.remove(edge)
        edge.end.edges.remove(edge)

    def render(self, terminal):
        for v in self.vertices:
            v.render(terminal)
        for e in self.edges:
            e.render(terminal)

    def get_layout(self):
        """Returns a list of (x, y, character)
        """
        layout = []
        for edge in self.edges:
            layout += edge.get_layout()
        for vertex in self.vertices:
            layout += vertex.get_layout()
        return layout

    def get_agents(self):
        "Return agents corresponding to this Graph, for use in-game."
        agents = [v.get_agent() for v in self.vertices]
        for e in self.edges:
            agents += e.get_agents()
        return agents

class Vertex:
    CHARACTERS = {
        "0000": " ",
        "0001": "═",
        "0010": "║",
        "0011": "╗",
        "0100": "═",
        "0101": "═",
        "0110": "╔",
        "0111": "╦",
        "1000": "║",
        "1001": "╝",
        "1010": "║",
        "1011": "╣",
        "1100": "╚",
        "1101": "╩",
        "1110": "╠",
        "1111": "╬",
    }
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.edges = []

    def __str__(self):
        return f"({self.x}, {self.y})"

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def neighbors(self):
        vertices = []
        for edge in self.edges:
            if self == edge.begin:
                vertices.append(edge.end)
            else:
                vertices.append(edge.begin)
        return vertices

    def render(self, terminal):
        print(terminal.move_xy(self.x, self.y) + self.get_character())

    def get_agent(self):
        return GraphAgent(self.get_character(), (self.x, self.y))

    def get_character(self):
        u = self.has_up_edge()
        r = self.has_right_edge()
        d = self.has_down_edge()
        l = self.has_left_edge()
        code = ''.join([str(int(direction)) for direction in [u, r, d, l]])
        return self.CHARACTERS[code]

    def get_layout(self):
        return [(self.x, self.y, self.get_character())]

    def has_up_edge(self):
        return any([v.x == self.x and v.y < self.y for v in self.neighbors()])

    def has_right_edge(self):
        return any([v.y == self.y and self.x < v.x for v in self.neighbors()])

    def has_down_edge(self):
        return any([v.x == self.x and self.y < v.y for v in self.neighbors()])

    def has_left_edge(self):
        return any([v.y == self.y and v.x < self.x for v in self.neighbors()])

class Edge:
    def __init__(self, begin, end):
        if not isinstance(begin, Vertex) or not isinstance(end, Vertex):
            raise ValueError("Tried to initialize an Edge with a non-vertex")
        if begin.x < end.x or begin.y < end.y:
            self.begin = begin
            self.end = end
        else:
            self.begin = end
            self.end = begin
        if not (self.is_horizontal() or self.is_vertical()):
            raise ValueError("Edges must be horizontal or vertical.")
        if self.is_horizontal() and self.is_vertical():
            raise ValueError("Self-edges are not allowed.")
        self.begin.edges.append(self)
        self.end.edges.append(self)

    def __str__(self):
        return f"{self.begin} -> {self.end}"

    def render(self, terminal):
        if self.is_horizontal():
            with terminal.location(self.begin.x + 1, self.begin.y):
                line = "═" * (self.end.x - self.begin.x - 1)
                print(line)
        else:
            for y in range(self.begin.y + 1, self.end.y):
                print(terminal.move_xy(self.begin.x, y) + "║")

    def get_layout(self):
        if self.is_horizontal():
            return [(x, self.begin.y, "=") for x in range(self.begin.x, self.end.x)]
        else:
            return [(self.begin.x, y, "║") for y in range(self.begin.y, self.end.y)]

    def get_agents(self):
        if self.is_horizontal():
            return [GraphAgent("=", (i, self.begin.y)) for i in range(self.begin.x + 1, self.end.x)]
        else:
            return [GraphAgent("║", (self.begin.x, j)) for j in range(self.begin.y + 1, self.end.y)]

    def is_horizontal(self):
        return self.begin.y == self.end.y

    def is_vertical(self):
        return self.begin.x == self.end.x

    def crosses(self, x, y):
        if self.is_horizontal():
            return self.begin.y == y and self.begin.x < x and x < self.end.x
        else:
            return self.begin.x == x and self.begin.y < y and y < self.end.y

    def remove(self):
        self.begin.edges.remove(self)
        self.end.edges.remove(self)

class GraphAgent:
    """An agent class for graph characters.
    """
    def __init__(self, character, position):
        self.character = character
        self.position = position

    def __repr__(self):
        return f"<GraphAgent {self.character} {self.position}>"



class GameError(Exception):
    pass

class AgentWithNameAlreadyExists(GameError):
    def __init__(self, name):
        message = f"There is already an agent named {agent.name} in the game"
        super().__init__(message)

class AgentNotFoundByName(GameError):
    def __init__(self, name):
        message = f"There is no agent named {name} in the game"
        super().__init__(message)

class AgentNotInGame(GameError):
    def __init__(self, agent):
        name = getattr(agent, "name", f"{agent.__class__.__name__}")
        message = f"Agent {name} is not in the game"
        super().__init__(message)

class AgentAlreadyInGame(GameError):
    def __init__(self, agent):
        name = getattr(agent, "name", f"{agent.__class__.__name__}")
        message = f"Cannot add agent {name} to the game; the agent is already in the game"
        super().__init__(message)

class IllegalMove(GameError):
    def __init__(self, agent, position):
        name = getattr(agent, "name",  f"{agent.__class__.__name__}")
        message = f"Agent {name} tried to move to {position}"
        super().__init__(message)

class AgentCharacterTooLarge(GameError):
    def __init__(self, agent, size, char_size):
        name = getattr(agent, "name", agent.__class__.__name__)
        message = (
            f"Agent {name} has size {size} but its character grid is "
            f"{char_size[0]} wide and {char_size[1]} tall, "
            f"which exceeds its declared size."
        )
        super().__init__(message)

class GraphError(GameError):
    pass

class TerminalTooSmall(GameError):
    BORDER_X = 2
    BORDER_Y = 3
    STATE_HEIGHT = 5

    def __init__(self, width=None, width_needed=None, height=None, height_needed=None):
        if width is not None and width_needed is not None and width < width_needed:
            err = f"The terminal width ({width}) is less than the required {width_needed}."
            super().__init__(err)
        elif height is not None and height_needed is not None and height < height_needed:
            err = f"The terminal height ({height}) is less than the required {height_needed}."
            super().__init__(err)
        else:
            raise ValueError(f"TerminalTooSmall called with illegal values.")



from retro.game import Game
from retro.examples.labyrinth.goal import Goal
from retro.examples.labyrinth.maze import Maze
from retro.examples.labyrinth.player import Player

WIDTH = 400
HEIGHT = 200
VIEW_WIDTH = 40
VIEW_HEIGHT = 20

state = {}
maze = Maze(WIDTH, HEIGHT)
start, end = maze.choose_start_and_end()
player = Player(start)
goal = Goal(end)
game = Game(
    [player] + maze.get_agents(), 
    state, 
    board_size=(WIDTH, HEIGHT), 
    view_size=(VIEW_WIDTH, VIEW_HEIGHT),
)
player.center_view(game)
game.play()



import json
from collections import defaultdict
from signal import signal, SIGWINCH
from time import sleep, perf_counter
from blessed import Terminal
from retro.view import View
from retro.change_dict import ChangeDict
from retro.agent import Tombstone
from retro.validation import (
    validate_agent, 
    validate_state,
    validate_agent_name,
    validate_position,
)
from retro.errors import (
    AgentWithNameAlreadyExists,
    AgentNotFoundByName,
    AgentAlreadyInGame,
    IllegalMove,
)

def agent_occupied_positions(agent):
    """Returns all board positions occupied by an agent."""
    x, y = agent.position
    size = getattr(agent, 'size', None)
    if size is None:
        return [(x, y)]
    w, h = size
    return [(x + dx, y + dy) for dy in range(h) for dx in range(w)]


class Game:
    """
    Creates a playable game.
    You will use Game to create games, but don't need to read or understand how
    this class works. The main work in creating a 

    Arguments: 
        agents (list): A list of agents to add to the game. 
        state (dict): A dict containing the game's initial state. 
        board_size (int, int): (Optional, default `(64, 32)`) The two-dimensional size 
            of the game board.
        view_size (int, int): (Optional, default is `board_size`) The two-dimensional size
            of the view. 
        view_position (int, int): (Optional) The (x, y) coordinates of the top left corner
            of the view. By default, this is (0, 0).
        debug (bool): (Optional) Turn on debug mode, showing log messages while playing.
        framerate (int): (Optional) The target number of frames per second at which the 
            game should run.
        color (str): (Optional) The game's background color scheme. `Available colors <https://blessed.readthedocs.io/en/latest/colors.html>`_.
        wait_for_enter (bool): (Optional) If True, the game screen stays open after the game ends until Enter or Escape is pressed. Defaults to False.
        dump_state (str): (Optional) A filename. If provided, the game state will be saved to that file as JSON when the game ends.

    ::

        # This example will create a simple game.
        from retro.game import Game
        from retro.agent import ArrowKeyAgent

        agents = [ArrowKeyAgent()]
        state = {}
        game = Game(agents, state)
        game.play()
        
    """
    STATE_HEIGHT = 5
    EXIT_CHARACTERS = ("KEY_ENTER", "KEY_ESCAPE")

    @property
    def view_position(self):
        return self._view_position

    @view_position.setter
    def view_position(self, position):
        if position != self.view_position:
            self._view_position = position
            self.view_position_changed = True

    def __init__(self, agents, state, board_size=(64, 32), view_size=None,
            view_position=(0, 0), debug=False, framerate=24, color="white_on_black",
            wait_for_enter=False, dump_state=None):
        self.log_messages = []
        self.agents_by_name = {}
        self.agents = []
        validate_state(state)
        self.state = ChangeDict(state)
        self.board_size = board_size
        self.view_size = view_size or board_size
        self._view_position = view_position
        self.view_position_changed = True
        self.debug = debug
        self.framerate = framerate
        self.turn_number = 0
        self.color = color
        self.wait_for_enter = wait_for_enter
        self.dump_state = dump_state
        self._position_cache = None
        for agent in agents:
            self.add_agent(agent)

    def play(self):
        """Starts the game.
        """
        self.playing = True
        terminal = Terminal()
        with terminal.fullscreen(), terminal.hidden_cursor(), terminal.cbreak():
            view = View(terminal, color=self.color)
            self.agent_positions = {}
            self.state.changed = True
            while self.playing:
                turn_start_time = perf_counter()
                self.turn_number += 1
                self.keys_pressed = self.collect_keystrokes(terminal)
                if self.debug and self.keys_pressed:
                    self.log("Keys: " + ', '.join(k.name or str(k) for k in self.keys_pressed))
                self.prior_view_position = self.view_position
                self.prior_agent_positions = self.agent_positions
                for agent in self.agents:
                    if hasattr(agent, 'handle_keystroke'):
                        for key in self.keys_pressed:
                            agent.handle_keystroke(key, self)
                    if hasattr(agent, 'play_turn'):
                        agent.play_turn(self)
                        self._position_cache = None
                    if getattr(agent, 'display', True):
                        for pos in agent_occupied_positions(agent):
                            if not self.on_board(pos):
                                raise IllegalMove(agent, pos)
                self.agent_positions = self.get_agents_by_position()
                view.render(self)
                self.state.changed = False
                self.view_position_changed = False
                turn_end_time = perf_counter()
                time_elapsed_in_turn = turn_end_time - turn_start_time
                time_remaining_in_turn = max(0, 1/self.framerate - time_elapsed_in_turn)
                sleep(time_remaining_in_turn)
            if self.dump_state:
                with open(self.dump_state, 'w') as f:
                    json.dump(dict(self.state), f)
            if self.wait_for_enter:
                while True:
                    if terminal.inkey().name in self.EXIT_CHARACTERS:
                        break

    def collect_keystrokes(self, terminal):
        keys = set()
        while True:
            key = terminal.inkey(0.001)
            if key: 
                keys.add(key)
            else:
                break
        return keys

    def log(self, message):
        """Write a log message. 
        Log messages are only shown when debug mode is on. 
        They can be very useful for debugging.

        Arguments:
            message (str): The message to log.
        """
        self.log_messages.append((self.turn_number, message))

    def end(self):
        """Ends the game. No more turns will run.
        """
        self.playing = False

    def add_agent(self, agent):
        """Adds an agent to the game.
        Whenever you want to add a new agent during the game, you must add it to 
        the game using this method.

        Arguments: 
            agent: An instance of an agent class. 
        """
        validate_agent(agent)
        if agent in self.agents:
            raise AgentAlreadyInGame(agent)
        if getattr(agent, "display", True):
            for pos in agent_occupied_positions(agent):
                if not self.on_board(pos):
                    raise IllegalMove(agent, pos)
        if hasattr(agent, "name"): 
            if agent.name in self.agents_by_name:
                raise AgentWithNameAlreadyExists(agent.name)
            self.agents_by_name[agent.name] = agent
        self.agents.append(agent)

    def get_agent_by_name(self, name):
        """Looks up an agent by name. 
        This is useful when one agent needs to interact with another agent.

        Arguments: 
            name (str): The agent's name. If there is no agent with this name, 
                you will get an error.

        Returns: 
            An agent.
        """
        validate_agent_name(name)
        if name in self.agents_by_name:
            return self.agents_by_name[name]
        else:
            raise AgentNotFoundByName(name)

    def is_empty(self, position):
        """Checks whether a position is occupied by any agents.

        Arguments: 
            position (int, int): The position to check.

        Returns: 
            A bool
        """
        return position not in self.get_agents_by_position()

    def get_agents_by_position(self):
        """Returns a dict where each key is a position (e.g. (10, 20)) and
        each value is a list containing all the agents at that position.
        This is useful when an agent needs to find out which other agents are
        on the same space or nearby.

        When called during ``play_turn()``, the dict reflects the board state
        at the start of the current agent's turn, including any moves made by
        agents that have already taken their turn this round.
        """
        if self._position_cache is not None:
            return self._position_cache
        positions = defaultdict(list)
        for agent in self.agents:
            if getattr(agent, "display", True):
                validate_position(agent.position)
                for pos in agent_occupied_positions(agent):
                    positions[pos].append(agent)
        self._position_cache = positions
        return positions

    def on_view(self, position):
        """Checks whether a position is within the current view.

        Arguments:
            position (int, int): The position to check.

        Returns:
            A bool
        """
        validate_position(position)
        x, y = position
        vox, voy = self.view_position
        vw, vh = self.view_size
        return vox <= x < vox + vw and voy <= y < voy + vh

    def remove_agent(self, agent):
        """Removes an agent from the game. 

        Arguments:
            agent (Agent): the agent to remove.
        """
        if agent not in self.agents:
            raise AgentNotInGame(agent)
        else:
            self.agents.remove(agent)
            if hasattr(agent, "name"):
                self.agents_by_name.pop(agent.name)

    def remove_agent_by_name(self, name):
        """Removes an agent from the game. 

        Arguments:
            name (str): the agent's name.
        """
        validate_agent_name(name)
        if name not in self.agents_by_name:
            raise AgentNotFoundByName(name)
        agent = self.agents_by_name.pop(name)
        self.agents.remove(agent)

    def on_board(self, position):
        """Checks whether a position is on the game board.

        Arguments: 
            position (int, int): The position to check

        Returns: 
            A bool
        """
        validate_position(position)
        x, y = position
        bx, by = self.board_size
        return x >= 0 and x < bx and y >= 0 and y < by






def add(vec0, vec1):
    """Adds two vectors.
    Got tired of doing this by hand. 
    """
    x0, y0 = vec0
    x1, y1 = vec1
    return (x0 + x1, y0 + y1)

def get_occupant(game, position):
    """Returns the agent at position, if there is one. 
    This function slightly simplifies the process of getting the 
    agent at a position: the game returns a list of agents at a position, 
    because some games allow more than one agent at a position.
    """
    positions_with_agents = game.get_agents_by_position()
    if position in positions_with_agents:
        agents_at_position = positions_with_agents[position]
        return agents_at_position[0]

def distance(vec0, vec1):
    """Returns the distance between two vectors, using the
    "manhattan distance," or the distance if you can only 
    move in the x-direction or the y-direction, but not 
    diagonally. Just like walking blocks in Manhattan :)
    """
    x0, y0 = vec0
    x1, y1 = vec1
    return abs(x1 - x0) + abs(y1 - y0)

